#include <linux/in.h>
#include <linux/net.h>
#include <linux/time.h>
#include <linux/if_ether.h>
#include <asm/errno.h>
#include <asm/unistd.h>
#include <asm/fcntl.h>

#undef NANOSEC_DELAY /* define NANOSEC_DELAY if you have problems */

/* Uninitialized globals */
static int errno;
static union {
  unsigned short sbuf[48];
  unsigned char cbuf[96];
} buf;

/* Macros */
#define __hns(X) ((((X)&0xff)<<8)|((X)>>8))
#define hhex(X,Y) ((Y=((X)>>4))>9?Y+'a'-10:Y+'0')
#define lhex(X,Y) ((Y=((X)&15))>9?Y+'a'-10:Y+'0')

/* Should the MSS setting be dropped?  need to get
 * this all reliable before I can test the effect
 */

/* Initialized globals */
static unsigned short obuf[24] = {
  /* the kernel does much of this for us */
  0x1045,                /* TOS, Vers, IHL */
  0,                     /* Total len */
  0,                     /* ID */
  0,                     /* Frag */
  0x0640,                /* TCP, TTL */
  0,                     /* Checksum */
  __hns(SRCIP>>16),      /* SRC IP */
  __hns(SRCIP&0xffff),   /* SRC IP */
  __hns(CLASSB),         /* DST IP */
  0,                     /* DST IP */
  __hns(SRCPORT),        /* SRC PRT */
  __hns(DSTPORT),        /* DST PRT */
  0,                     /* Initial seq */
  0,                     /* Initial seq */
  0,                     /* ACK seq */
  0,                     /* ACK seq */
  0x0260,                /* SYN, size */
  0x0002,                /* Window size */
  0,                     /* Checksum - this one we need to fill */
  0,                     /* Urgent pointer */
  0x0402,                /* Set MSS */
  0x0002,                /* MSS */
  /* pseudo-header fields (some are before the tcp section) */
  0x0600,                /* TCP */
  0x1800                 /* TCP len */
};
static union {
  struct sockaddr sa;
  struct sockaddr_in sa_in;
} sa = {{ AF_INET, "eth0" }};
static int fdmax = 1;
static union {
  struct timeval tv;
  struct timespec ts;
} tm = {{0, 0}};
static fd_set fds = {{0}};
static unsigned long rargs[4] = { 0, (unsigned long)&buf, sizeof(buf.cbuf), 0 };
static unsigned long sargs1[3] = { PF_INET, SOCK_PACKET, __hns(ETH_P_ALL) };
static unsigned long sargs2[3] = { PF_INET, SOCK_RAW, IPPROTO_RAW };
static unsigned long bargs[3] = { 0, (unsigned long)&sa, sizeof(struct sockaddr) };
static unsigned long ssoargs[5] = { 0, IPPROTO_IP, IP_HDRINCL, (unsigned long)&fdmax, 1 };
static unsigned long stoargs[6] = { 0, (unsigned long)&obuf, 44, 0, (unsigned long)&sa, sizeof(struct sockaddr_in) };
static unsigned char minibuf[5] = { ' ', ' ', ' ', ' ', '\n' };

/* Syscalls */
static __inline__ _syscall5(int,_newselect,int,a1,fd_set *,a2,fd_set *,a3,fd_set *,a4,struct timeval *,a5)
static __inline__ _syscall3(int,write,int,a1,const char *,a2,off_t,a3)
static __inline__ _syscall2(int,socketcall,int,a1,unsigned long *,args)
static __inline__ _syscall1(int,exit,int,a1)
static __inline__ _syscall2(int,nanosleep,struct timespec *,a1,struct timespec *,a2)
static __inline__ _syscall3(long,fcntl,int,a1,int,a2,unsigned long,a3)

/* Functions */
/* GCC optimizes better if things are split into chunks like this */

static __inline__ unsigned long pcsum() {
  unsigned short res = 0, res0;
  int i;

  for(i = 6; i<24; i++) {
    res0 = res;
    res += ntohs(obuf[i]);
    if (res < res0)
      res++;
  }
  return res;
}

static __inline__ void quick_write(unsigned char net, unsigned char host) {
  unsigned char j;

  minibuf[0] = hhex(net, j);
  minibuf[1] = lhex(net, j);
  minibuf[2] = hhex(host, j);
  minibuf[3] = lhex(host, j);
  write(1, (char *)minibuf, 5);
}

static __inline__ int get_packets() {
  int i;
  unsigned char j;

  for(;;) {
    FD_SET(fdmax, &fds);
    while ((i=_newselect(fdmax+1, &fds, NULL, NULL, &tm.tv)) < 0 && errno == EINTR);
    if(i < 0)
      return 1;
    if(i == 0)
      return 0;
    while ((i=socketcall(SYS_RECV, &rargs[0])) < 0 && (errno == EINTR || errno == EWOULDBLOCK));
    if(i < 0)
      return 1;
    /* Doing these two in one hits some kind of gcc bug (?) */
    j = buf.cbuf[14];
    j &= 0x0f;
    /* own address matching doesn't have to be exact - if the packet fits... */
    if(buf.sbuf[2]!=__hns(MACADDR3) ||    /* to me */
       buf.sbuf[13]!=__hns(CLASSB) ||     /* scanned net */
       buf.sbuf[6]!=__hns(0x0800) ||      /* ip */
       buf.cbuf[23]!=0x06 ||              /* tcp */
       buf.sbuf[7+2*j]==__hns(DSTPORT) || /* port */
       buf.cbuf[27+4*j]!=0x12)            /* syn+ack */
      continue;
    quick_write(buf.cbuf[28], buf.cbuf[29]);
  }
}

static __inline__ int fast_get_packets() {
  int i;
  unsigned char j;

#ifdef NANOSEC_DELAY
  /* with these 2 lines, it's really slow.  without them, it locks up
   * and misses lots of ack's on some systems.
   */
  tm.ts.tv_nsec = 1;
  while (nanosleep(&tm.ts,&tm.ts) < 0 && errno == EINTR);
#endif
  for (;;) {
    while ((i=socketcall(SYS_RECV, &rargs[0])) < 0 && errno == EINTR);
    if(i < 0)
      return errno != EWOULDBLOCK;
    /* Doing these two in one hits some kind of gcc bug (?) */
    j = buf.cbuf[14];
    j &= 0x0f;
    /* own address matching doesn't have to be exact - if the packet fits... */
    if(buf.sbuf[2]!=__hns(MACADDR3) ||    /* to me */
       buf.sbuf[13]!=__hns(CLASSB) ||     /* scanned net */
       buf.sbuf[6]!=__hns(0x0800) ||      /* ip */
       buf.cbuf[23]!=0x06 ||              /* tcp */
       buf.sbuf[7+2*j]!=__hns(DSTPORT) || /* port */
       buf.cbuf[27+4*j]!=0x12)            /* syn+ack */
      continue;
    quick_write(buf.cbuf[28], buf.cbuf[29]);
  }
}

static __inline__ int send_packets() {
  unsigned short h = 0, partsum;

  partsum = pcsum();
  for(h=0;h!=0xffff;) {
    h++;
    partsum++;
    if(partsum == 0)
      partsum = 1;
    if((unsigned char)((h+1)&0xff) <= 1)  
      continue;
    obuf[18] = ~htons(partsum);
    *(((unsigned short *)&sa.sa_in.sin_addr.s_addr)+1) = obuf[9] = ntohs(h);
    if(socketcall(SYS_SENDTO, stoargs) != 44 || fast_get_packets())
      return 1;
  }
  return 0;
}

extern void _start() {
  if((ssoargs[0] = stoargs[0] = socketcall(SYS_SOCKET, &sargs2[0]))<0||
     socketcall(SYS_SETSOCKOPT, ssoargs)<0||
     (rargs[0] = bargs[0] = fdmax = socketcall(SYS_SOCKET, &sargs1[0]))<0||
     socketcall(SYS_BIND, &bargs[0])<0 ||
     fcntl(fdmax, F_SETFL, O_RDONLY|O_NONBLOCK)<0)
    goto ERR_EXIT;
  sa.sa_in.sin_port = __hns(DSTPORT);
  sa.sa_in.sin_addr.s_addr = __hns(CLASSB);
  if(send_packets())
    goto ERR_EXIT;
  tm.tv.tv_sec = TIMEOUT;
  if(!get_packets())
    goto EXIT;
ERR_EXIT:
  write(1, "XXXX\n", 5);
EXIT:
  exit(0);
}
